package cn.dankal.test.concurrent;

import cn.dankal.test.common.Const;
import cn.dankal.test.factory.ThreadExecutorFactory;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.StringSerializer;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * Kafka 测试: 多线程环境的 Kafka Producer
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-19
 */
public class KafkaProducerOnConcurrentTest {

    /**
     * ThreadPoolExecutor
     */
    private static ThreadPoolExecutor threadPoolExecutor;

    /**
     * KafkaProducer
     */
    private static KafkaProducer<String, String> kafkaProducer;

    @BeforeAll
    public static void before() {
        threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("dankal-test-kafka");

        Properties properties = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        properties.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 序列化器
        properties.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());
        // message value 序列化器
        properties.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());

        // 通过 Properties 配置 实例化 KafkaProducer
        kafkaProducer = new KafkaProducer<>(properties);
    }

    /**
     * 测试多线程环境下的 Kafka Producer
     */
    @Test
    public void test() throws InterruptedException {
        final int threadCount = 5;
        // 保证所有线程全部完成消息发送
        final CountDownLatch countDownLatch = new CountDownLatch(threadCount);
        // 保证 所有线程在同一起跑线
        final CyclicBarrier cyclicBarrier = new CyclicBarrier(threadCount);
        for (int i = 0; i < threadCount; i++) {
            final int index = i + 1;
            Runnable kafkaProducerRunnable = () -> {
                final String key = "dankal-test-" + index;
                final String message = "dankal-test-kafka-producer-on-concurrent";
                ProducerRecord<String, String> producerRecord = new ProducerRecord<>(
                        Const.Kafka.Topic.HELLO_TOPIC,
                        key,
                        message
                );
                try {
                    cyclicBarrier.await();
                } catch (Exception e) {
                    // ignored
                }
                kafkaProducer.send(
                        producerRecord,
                        (recordMetadata, exception) -> {
                            countDownLatch.countDown();
                            if (null != exception) {
                                // 发送异常
                                exception.printStackTrace();
                                return;
                            }
                            // 发送成功
                            System.out.printf(
                                    "完成消息发送 %d ==> 主题: %s  分区: %s 偏移量: %d%n",
                                    recordMetadata.timestamp(),
                                    recordMetadata.topic(),
                                    recordMetadata.partition(),
                                    recordMetadata.offset());
                        });
            };
            threadPoolExecutor.submit(kafkaProducerRunnable);
        }
        // 等待所有消息执行完成
        countDownLatch.await();
    }


    @AfterAll
    public static void after() {
        if (Objects.nonNull(kafkaProducer)) {
            kafkaProducer.close();
        }
        if (null != threadPoolExecutor) {
            threadPoolExecutor.shutdown();
        }
    }
}
