package cn.dankal.test.consumercommit;

import cn.dankal.test.common.Const;
import cn.hutool.core.collection.CollectionUtil;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Properties;

/**
 * 测试Kafka: Kafka Consumer 偏移量提交方式
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class KafkaConsumerCommitOffsetTest {


    /**
     * KafkaConsumer
     */
    private KafkaConsumer<String, String> kafkaConsumer;

    /**
     * Properties
     */
    private static Properties properties;


    @BeforeAll
    public static void before() {
        properties = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        properties.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 反序列化器
        properties.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());
        // message value 反序列化器
        properties.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());

        properties.put(ConsumerConfig.GROUP_ID_CONFIG, Const.Kafka.ConsumerGroupId.HELLO_CONSUMER_GROUP_ID);
    }

    /**
     * 测试 Kafka Consumer 自动提交偏移量
     */
    @Test
    public void testAutoCommitOffset() {
        // 通过 Properties 配置 实例化 KafkaConsumer
        kafkaConsumer = new KafkaConsumer<>(properties);
        // consumer 订阅 主题
        kafkaConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        while (true) {
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                // 成功拉取 ConsumerRecords 后,5000毫秒后 Consumer 会自动提交偏移量
                System.out.printf(
                        "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                        Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                        consumerRecord.topic(),
                        consumerRecord.partition(),
                        consumerRecord.offset(),
                        consumerRecord.key(),
                        consumerRecord.value());
            }
        }
    }


    /**
     * 测试 Kafka Consumer 同步手动提交
     */
    @Test
    public void testSyncCommitOffset() {
        properties.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        // 通过 Properties 配置 实例化 KafkaConsumer
        kafkaConsumer = new KafkaConsumer<>(properties);
        // consumer 订阅 主题
        kafkaConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        while (true) {
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                // 成功拉取 ConsumerRecords 后,5000毫秒后 Consumer 会自动提交偏移量
                System.out.printf(
                        "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                        Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                        consumerRecord.topic(),
                        consumerRecord.partition(),
                        consumerRecord.offset(),
                        consumerRecord.key(),
                        consumerRecord.value());
            }
            if (CollectionUtil.isNotEmpty(consumerRecords)) {
                // 同步手动提交偏移量
                kafkaConsumer.commitSync();
                System.out.println("after invoke commitSync()");
            }
        }
    }


    /**
     * 测试 Kafka Consumer 异步手动提交
     */
    @Test
    public void testAsyncCommitOffset() {
        properties.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        // 通过 Properties 配置 实例化 KafkaConsumer
        kafkaConsumer = new KafkaConsumer<>(properties);
        // consumer 订阅 主题
        kafkaConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        while (true) {
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                // 成功拉取 ConsumerRecords 后,5000毫秒后 Consumer 会自动提交偏移量
                System.out.printf(
                        "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                        Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                        consumerRecord.topic(),
                        consumerRecord.partition(),
                        consumerRecord.offset(),
                        consumerRecord.key(),
                        consumerRecord.value());
            }
            if (CollectionUtil.isNotEmpty(consumerRecords)) {
                // 异步手动提交偏移量
                kafkaConsumer.commitAsync((offsets, exception) -> {
                    if (null != exception) {
                        // 提交偏移量发生异常
                        exception.printStackTrace();
                        return;
                    }
                    // 成功提交偏移量
                    for (Map.Entry<TopicPartition, OffsetAndMetadata> topicPartitionOffsetAndMetadataEntry : offsets.entrySet()) {
                        TopicPartition topicPartition = topicPartitionOffsetAndMetadataEntry.getKey();
                        OffsetAndMetadata offsetAndMetadata = topicPartitionOffsetAndMetadataEntry.getValue();
                        System.out.printf(
                                "成功提交偏移量. 主题: %s 分区:%s 偏移量: %d%n",
                                topicPartition.topic(),
                                topicPartition.partition(),
                                offsetAndMetadata.offset()
                        );
                    }

                });
                System.out.println("after invoke commitASync(OffsetCommitCallback)");
            }
        }
    }


    /**
     * 测试 Kafka Consumer 混合式提交(commitASync/commitSync)
     */
    @Test
    public void testKafkaMixtureCommitOffset() {
        properties.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        // 通过 Properties 配置 实例化 KafkaConsumer
        kafkaConsumer = new KafkaConsumer<>(properties);
        // consumer 订阅 主题
        kafkaConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        while (true) {
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            try {
                for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                    // 成功拉取 ConsumerRecords 后,5000毫秒后 Consumer 会自动提交偏移量
                    System.out.printf(
                            "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                            Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                            consumerRecord.topic(),
                            consumerRecord.partition(),
                            consumerRecord.offset(),
                            consumerRecord.key(),
                            consumerRecord.value());
                }
                if (CollectionUtil.isNotEmpty(consumerRecords)) {
                    // 异步手动提交偏移量
                    kafkaConsumer.commitAsync((offsets, exception) -> {
                        if (null != exception) {
                            // 提交偏移量发生异常
                            exception.printStackTrace();
                            return;
                        }
                        // 成功提交偏移量
                        for (Map.Entry<TopicPartition, OffsetAndMetadata> topicPartitionOffsetAndMetadataEntry : offsets.entrySet()) {
                            TopicPartition topicPartition = topicPartitionOffsetAndMetadataEntry.getKey();
                            OffsetAndMetadata offsetAndMetadata = topicPartitionOffsetAndMetadataEntry.getValue();
                            System.out.printf(
                                    "成功提交偏移量. 主题: %s 分区:%s 偏移量: %d%n",
                                    topicPartition.topic(),
                                    topicPartition.partition(),
                                    offsetAndMetadata.offset()
                            );
                        }

                    });
                    System.out.println("after invoke commitASync(OffsetCommitCallback)");
                }
            } finally {
                if (null != kafkaConsumer && CollectionUtil.isNotEmpty(consumerRecords)) {
                    kafkaConsumer.commitSync();
                    System.out.println("finally use commitSync() commit offset");
                }
            }
        }
    }


    /**
     * 测试 kafka Consumer: 特定提交
     */
    @Test
    public void testKafkaSpecificCommitOffset() {
        properties.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        // 通过 Properties 配置 实例化 KafkaConsumer
        kafkaConsumer = new KafkaConsumer<>(properties);
        Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap = new HashMap<>(1 << 5);
        // 提交频率
        final int commitFrequency = 2;
        // consumer 订阅 主题
        kafkaConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        while (true) {
            int count = 0;
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                try {
                    // 成功拉取 ConsumerRecords 后,5000毫秒后 Consumer 会自动提交偏移量
                    System.out.printf(
                            "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                            Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                            consumerRecord.topic(),
                            consumerRecord.partition(),
                            consumerRecord.offset(),
                            consumerRecord.key(),
                            consumerRecord.value());

                    offsetAndMetadataMap.put(
                            new TopicPartition(consumerRecord.topic(), consumerRecord.partition()),
                            new OffsetAndMetadata(consumerRecord.offset() + 1, "no metadata")
                    );
                    if (count % commitFrequency == 0) {
                        kafkaConsumer.commitAsync(offsetAndMetadataMap, (offsets, exception) -> {
                            if (null != exception) {
                                // 提交偏移量发生异常
                                exception.printStackTrace();
                                return;
                            }
                            // 成功提交偏移量
                            for (Map.Entry<TopicPartition, OffsetAndMetadata> topicPartitionOffsetAndMetadataEntry : offsets.entrySet()) {
                                TopicPartition topicPartition = topicPartitionOffsetAndMetadataEntry.getKey();
                                OffsetAndMetadata offsetAndMetadata = topicPartitionOffsetAndMetadataEntry.getValue();
                                System.out.printf(
                                        "成功提交偏移量. 主题: %s 分区:%s 偏移量: %d%n",
                                        topicPartition.topic(),
                                        topicPartition.partition(),
                                        offsetAndMetadata.offset()
                                );
                            }
                        });
                    }
                    count++;
                } finally {
                    if (null != kafkaConsumer) {
                        kafkaConsumer.commitSync();
                    }
                }
            }
        }
    }

    @AfterEach
    public void after() {
        if (Objects.nonNull(kafkaConsumer)) {
            kafkaConsumer.close();
        }
    }
}
