package cn.dankal.test.consumergroup;

import cn.dankal.test.common.Const;
import cn.dankal.test.factory.ThreadExecutorFactory;
import cn.hutool.core.collection.CollectionUtil;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.common.serialization.StringDeserializer;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.Properties;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * 测试 Kafka: 消费群组
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class KafkaConsumerGroupTest {

    /**
     * Properties
     */
    private Properties properties;

    /**
     * ThreadPoolExecutor
     */
    private final ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("dankal-test-kafka-consumer-group");

    @BeforeEach
    public void before() {
        this.properties = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        properties.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 反序列化器
        properties.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());
        // message value 反序列化器
        properties.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());
    }

    /**
     * 测试Kafka Consumer Group A
     */
    @Test
    public void testKafkaConsumerGroupA() throws InterruptedException {
        properties.put(ConsumerConfig.GROUP_ID_CONFIG, Const.Kafka.ConsumerGroupId.TEST_GROUP_A);
        threadPoolExecutor.submit(() -> {
            KafkaConsumer<String, String> groupAConsumer01 = new KafkaConsumer<>(properties);
            groupAConsumer01.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
            this.doConsumer("groupAConsumer01", groupAConsumer01);
        });
        threadPoolExecutor.submit(() -> {
            KafkaConsumer<String, String> groupAConsumer02 = new KafkaConsumer<>(properties);
            groupAConsumer02.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
            this.doConsumer("groupAConsumer02", groupAConsumer02);
        });
        threadPoolExecutor.submit(() -> {
            KafkaConsumer<String, String> groupAConsumer03 = new KafkaConsumer<>(properties);
            groupAConsumer03.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
            this.doConsumer("groupAConsumer03", groupAConsumer03);
        });
        // 不让程序终止运行
        new CountDownLatch(1).await();
    }

    /**
     * 测试Kafka Consumer Group B
     */
    @Test
    public void testKafkaConsumerGroupB() throws InterruptedException {
        properties.put(ConsumerConfig.GROUP_ID_CONFIG, Const.Kafka.ConsumerGroupId.TEST_GROUP_B);

        threadPoolExecutor.submit(() -> {
            KafkaConsumer<String, String> groupBConsumer01 = new KafkaConsumer<>(properties);
            groupBConsumer01.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
            this.doConsumer("groupBConsumer01", groupBConsumer01);
        });
        threadPoolExecutor.submit(() -> {
            KafkaConsumer<String, String> groupBConsumer02 = new KafkaConsumer<>(properties);
            groupBConsumer02.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
            this.doConsumer("groupBConsumer02", groupBConsumer02);
        });

        new CountDownLatch(1).await();
    }

    /**
     * 测试Kafka Consumer Group C
     */
    @Test
    public void testKafkaConsumerGroupC() {
        properties.put(ConsumerConfig.GROUP_ID_CONFIG, Const.Kafka.ConsumerGroupId.TEST_GROUP_C);
        KafkaConsumer<String, String> groupCConsumer = new KafkaConsumer<>(properties);
        groupCConsumer.subscribe(CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC));
        this.doConsumer("groupCConsumer", groupCConsumer);
    }

    /**
     * 执行 Kafka 消息消费
     *
     * @param consumerName  the  consumer name
     * @param kafkaConsumer the KafkaConsumer
     */
    private void doConsumer(String consumerName, KafkaConsumer<String, String> kafkaConsumer) {
        while (true) {
            ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
            for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                System.out.printf(
                        "%s : %s 消费消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                        Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                        consumerName,
                        consumerRecord.topic(),
                        consumerRecord.partition(),
                        consumerRecord.offset(),
                        consumerRecord.key(),
                        consumerRecord.value());
            }
        }

    }
}
