package cn.dankal.test.rebalance;

import cn.dankal.test.common.Const;
import cn.hutool.core.collection.CollectionUtil;
import org.apache.kafka.clients.consumer.ConsumerRebalanceListener;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndMetadata;
import org.apache.kafka.common.TopicPartition;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 测试 Kafka 分区再均衡: RebalanceHandler
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class RebalanceListener implements ConsumerRebalanceListener {

    /**
     * Map<TopicPartition, OffsetAndMetadata> 当前消费者分区与偏移量信息
     */
    private final Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap;

    /**
     * current ConsumerWorker KafkaConsumer instance
     */
    private final KafkaConsumer<String, String> kafkaConsumer;

    public RebalanceListener(Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap, KafkaConsumer<String, String> kafkaConsumer) {
        this.offsetAndMetadataMap = offsetAndMetadataMap;
        this.kafkaConsumer = kafkaConsumer;
    }


    /**
     * 再执行分区再平衡之前,回调此方法
     * <p>
     * Collection<TopicPartition>  ==> 当前 Consumer 所有拥有消费权的分区信息
     * <p>
     * 我们一般在此方法完成分区再平衡之前,分区偏移量的存储
     *
     * @param partitions Collection<TopicPartition>
     */
    @Override
    public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
        // 分区再均衡之前

        System.out.printf(
                "%s:onPartitionsRevoked 分区: %s%n",
                Thread.currentThread().getName(),
                partitions
        );
        System.out.printf(
                "%s 即将开始参与分区再均衡,即将提交工作偏移量.偏移量信息为: %s%n",
                Thread.currentThread().getName(),
                this.offsetAndMetadataMap
        );
        System.out.printf("偏移量分区信息:%s%n", Const.PARTITION_OFFSET_INFO_MAP);

        // TODO transaction  begin
        try {
            for (TopicPartition partition : partitions) {
                //TODO 将偏移量执行事务写入
                Const.PARTITION_OFFSET_INFO_MAP.put(
                        partition,
                        offsetAndMetadataMap.get(partition).offset());
            }

            //TODO transaction commit

            //可选 将偏移量信息提交到 Kafka 服务器
            //因为发生分区再均衡 分区偏移量是我们自己在维护 Const.PARTITION_OFFSET_INFO_MAP
            kafkaConsumer.commitSync(offsetAndMetadataMap);

        } catch (Exception e) {
            // TODO transaction rollback
            // TODO handler exception
            e.printStackTrace();
        }
    }

    /**
     * 完成分区再平衡之后,回调当前方法
     * <p>
     * Collection<TopicPartition>  ==> 当前 Consumer 所有拥有消费权的分区信息
     * <p>
     * 我们一般在此方法内,通过在方法{@link #onPartitionsRevoked(Collection)}存储的分区偏移量信息,对分区偏移量信息进行矫正
     *
     * @param partitions Collection<TopicPartition>
     */
    @Override
    public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
        System.out.printf(
                "%s:onPartitionsAssigned 分区: %s%n",
                Thread.currentThread().getName(),
                partitions
        );
        System.out.printf(
                "%s 已完成参与分区再均衡.偏移量信息为: %s%n",
                Thread.currentThread().getName(),
                Const.PARTITION_OFFSET_INFO_MAP
        );

        //分区偏移量矫正
        for (TopicPartition partition : partitions) {
            System.out.printf("%s:topic %n", partition);
            // 获取我们自己维护的分区偏移量信息
            Long offset = Const.PARTITION_OFFSET_INFO_MAP.get(partition);
            if (null == offset) {
                continue;
            }
            // 调整分区偏移量
            // 从特定偏移量开始记录分区偏移量
            // 保证分区再均衡之后,数据偏移量不会错乱
            kafkaConsumer.seek(partition, offset);
        }
        resetConsumerOffsetAndMetadataMap(partitions);
    }

    /**
     * 重置当前 Kafka Consumer 的 <code>Map<TopicPartition, OffsetAndMetadata></code>
     *
     * @param partitions Collection<TopicPartition>
     */
    private void resetConsumerOffsetAndMetadataMap(Collection<TopicPartition> partitions) {
        if (CollectionUtil.isEmpty(partitions)) {
            return;
        }
        List<Integer> ownPartitionIdList = partitions.stream()
                .map(TopicPartition::partition)
                .collect(Collectors.toList());
        System.out.printf("%s before: resetConsumerOffsetAndMetadataMap %s %n", Thread.currentThread().getName(), this.offsetAndMetadataMap);
        for (TopicPartition topicPartition : this.offsetAndMetadataMap.keySet()) {
            if (!ownPartitionIdList.contains(topicPartition.partition())) {
                // 分区再平衡时,当前消费者已经不再拥有当前 partition 的消费权
                // 避免出现覆盖,故需要在当前 Consumer 的 offsetAndMetadataMap 移除 当前 partition 的信息
                // 出现场景可以参阅项目目录:  test-kafka-rebalance-log-error.html
                this.offsetAndMetadataMap.remove(topicPartition);
            }
        }
        System.out.printf("%s after: resetConsumerOffsetAndMetadataMap %s %n", Thread.currentThread().getName(), this.offsetAndMetadataMap);
    }
}
