package cn.dankal.test.cloud.consumer.consumer;

import cn.dankal.test.cloud.consumer.common.Common;
import cn.dankal.test.cloud.consumer.common.Const;
import cn.dankal.test.cloud.consumer.entity.DankalSimpleMessage;
import cn.dankal.test.cloud.consumer.input.DankalMessageInput;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.annotation.StreamListener;
import org.springframework.integration.annotation.ServiceActivator;
import org.springframework.integration.context.IntegrationContextUtils;
import org.springframework.kafka.support.Acknowledgment;
import org.springframework.kafka.support.KafkaHeaders;
import org.springframework.messaging.Message;
import org.springframework.messaging.handler.annotation.Header;
import org.springframework.messaging.handler.annotation.Payload;
import org.springframework.messaging.support.ErrorMessage;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;
import java.util.Objects;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-02-01
 */
@Slf4j
@Component
public class DankalMessageConsumer {

    @StreamListener(DankalMessageInput.INPUT_NAME)
    public void onMessage(
            @Payload DankalSimpleMessage dankalSimpleMessage,
            @Header(KafkaHeaders.ACKNOWLEDGMENT) Acknowledgment acknowledgment) {
        log.info(
                "DankalMessageConsumer: {} {} 成功消费: {}",
                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                Common.getSpringKafkaConsumerName(Thread.currentThread().getName()),
                dankalSimpleMessage
        );
        if (dankalSimpleMessage.getId() % 2 == 0) {
            acknowledgment.acknowledge();
            log.info("{} 消费确认", dankalSimpleMessage.getId());
        }

        // 测试 spring-cloud-stream-kafka 消息重试
        //throw new RuntimeException("dankal test retry consumer message");
    }

    /**
     * 处理指定 {@link  org.springframework.cloud.stream.annotation.Input} 的异常
     * <pre>
     *   异常 input name :  Channel(<destination>.<group>.errors)
     *   举个栗子:
     *         {@link DankalMessageInput#dankalMessageInputChannel()}
     *              destination:  cloud-kafka-dankal-message-topic
     *              group: cloud-kafka-dankal-message-consumer-group
     *         则 此 input 对应消费异常管道为:
     *         cloud-kafka-dankal-message-topic.cloud-kafka-dankal-message-consumer-group.errors
     *
     * </pre>
     */
    @ServiceActivator(inputChannel = "cloud-kafka-dankal-message-topic.cloud-kafka-dankal-message-consumer-group.errors")
    public void handlerDankalMessageInputError(ErrorMessage errorMessage) {
        log.info("current error is  handler to DankalMessageInput ");
        log.error("payload error message: {}", errorMessage.getPayload().getMessage());
        Message<?> originalMessage = errorMessage.getOriginalMessage();
        if (Objects.nonNull(originalMessage)) {
            log.error("origin payload message:{}", originalMessage.getPayload());
        }
        log.error("error headers:{}", errorMessage.getHeaders());
    }

    /**
     * input 全局异常异常处理
     */
    @StreamListener(IntegrationContextUtils.ERROR_CHANNEL_BEAN_NAME)
    public void globalInputErrorHandler(ErrorMessage errorMessage) {
        log.info("current error is  input error handler ");
        log.error("payload error message: {}", errorMessage.getPayload().getMessage());
        Message<?> originalMessage = errorMessage.getOriginalMessage();
        if (Objects.nonNull(originalMessage)) {
            log.error("origin payload message:{}", originalMessage.getPayload());
        }
        log.error("error headers:{}", errorMessage.getHeaders());
    }

}
