# 蛋壳创意科技-Kafka负载再均衡示例小析

> company: 	[蛋壳创意科技](https://www.dankal.cn/)
>
> code:	[]()
>
> log: 	[test-kafka-rebalance-log](https://res.mercymodest.com/res/test-kafka-rebalance-log.html)

## 测试用例总览

**项目:	dankal-test-kafka-protogenesis**

**根目录:	test**

**包名:	cn.dankal.test.rebalance**

> 消息生产者:	ProducerWorkerTest
>
> 消息消费者:	ConsumerWorker
>
> 分区再均衡监听器:	RebalanceListener
>
> 测试类:	KafkaRebalanceTest

## 测试用例详情

### 前景小述

> 理论上，我们需要将分区偏移量信息保存到关系数据库,通过数据库事务控制操作原子性~然，为了简化测试用例，测试用例使用一个 `java.util.concurrent.ConcurrentHashMap`来存储 
>
> ![](https://img.mercymodest.com/public/20210127105302.png)
>
> 然后 以 `TODO`的形式,提供 `Transactional` 的参考

==测试用例是对如下图示的具体实现，所以说呢，阅码第一步，读图走起~==

![](https://img.mercymodest.com/public/20210126154516.jpg)

### RebalanceListener:  自定义再均衡监听器

```java
package cn.dankal.test.rebalance;

import cn.dankal.test.boot.common.Const;
import cn.hutool.core.collection.CollectionUtil;
import org.apache.kafka.clients.consumer.ConsumerRebalanceListener;
import org.apache.kafka.clients.consumer.KafkaConsumer;
import org.apache.kafka.clients.consumer.OffsetAndMetadata;
import org.apache.kafka.common.TopicPartition;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 测试 Kafka 分区再均衡: RebalanceHandler
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class RebalanceListener implements ConsumerRebalanceListener {

    /**
     * Map<TopicPartition, OffsetAndMetadata> 当前消费者分区与偏移量信息
     */
    private final Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap;

    /**
     * current ConsumerWorker KafkaConsumer instance
     */
    private final KafkaConsumer<String, String> kafkaConsumer;

    public RebalanceListener(Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap, KafkaConsumer<String, String> kafkaConsumer) {
        this.offsetAndMetadataMap = offsetAndMetadataMap;
        this.kafkaConsumer = kafkaConsumer;
    }


    /**
     * 再执行分区再平衡之前,回调此方法
     * <p>
     * Collection<TopicPartition>  ==> 当前 Consumer 所有拥有消费权的分区信息
     * <p>
     * 我们一般在此方法完成分区再平衡之前,分区偏移量的存储
     *
     * @param partitions Collection<TopicPartition>
     */
    @Override
    public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
        // 分区再均衡之前

        System.out.printf(
                "%s:onPartitionsRevoked 分区: %s%n",
                Thread.currentThread().getName(),
                partitions
        );
        System.out.printf(
                "%s 即将开始参与分区再均衡,即将提交工作偏移量.偏移量信息为: %s%n",
                Thread.currentThread().getName(),
                this.offsetAndMetadataMap
        );
        System.out.printf("偏移量分区信息:%s%n", Const.PARTITION_OFFSET_INFO_MAP);

        // TODO transaction  begin
        try {
            for (TopicPartition partition : partitions) {
                //TODO 将偏移量执行事务写入
                Const.PARTITION_OFFSET_INFO_MAP.put(
                        partition,
                        offsetAndMetadataMap.get(partition).offset());
            }

            //TODO transaction commit

            //可选 将偏移量信息提交到 Kafka 服务器
            //因为发生分区再均衡 分区偏移量是我们自己在维护 Const.PARTITION_OFFSET_INFO_MAP
            kafkaConsumer.commitSync(offsetAndMetadataMap);

        } catch (Exception e) {
            // TODO transaction rollback
            // TODO handler exception
            e.printStackTrace();
        }
    }

    /**
     * 完成分区再平衡之后,回调当前方法
     * <p>
     * Collection<TopicPartition>  ==> 当前 Consumer 所有拥有消费权的分区信息
     * <p>
     * 我们一般在此方法内,通过在方法{@link #onPartitionsRevoked(Collection)}存储的分区偏移量信息,对分区偏移量信息进行矫正
     *
     * @param partitions Collection<TopicPartition>
     */
    @Override
    public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
        System.out.printf(
                "%s:onPartitionsAssigned 分区: %s%n",
                Thread.currentThread().getName(),
                partitions
        );
        System.out.printf(
                "%s 已完成参与分区再均衡.偏移量信息为: %s%n",
                Thread.currentThread().getName(),
                Const.PARTITION_OFFSET_INFO_MAP
        );

        //分区偏移量矫正
        for (TopicPartition partition : partitions) {
            System.out.printf("%s:topic %n", partition);
            // 获取我们自己维护的分区偏移量信息
            Long offset = Const.PARTITION_OFFSET_INFO_MAP.get(partition);
            if (null == offset) {
                continue;
            }
            // 调整分区偏移量
            // 从特定偏移量开始记录分区偏移量
            // 保证分区再均衡之后,数据偏移量不会错乱
            kafkaConsumer.seek(partition, offset);
        }
        resetConsumerOffsetAndMetadataMap(partitions);
    }

    /**
     * 重置当前 Kafka Consumer 的 <code>Map<TopicPartition, OffsetAndMetadata></code>
     *
     * @param partitions Collection<TopicPartition>
     */
    private void resetConsumerOffsetAndMetadataMap(Collection<TopicPartition> partitions) {
        if (CollectionUtil.isEmpty(partitions)) {
            return;
        }
        List<Integer> ownPartitionIdList = partitions.stream()
                .map(TopicPartition::partition)
                .collect(Collectors.toList());
        System.out.printf("%s before: resetConsumerOffsetAndMetadataMap %s %n", Thread.currentThread().getName(), this.offsetAndMetadataMap);
        for (TopicPartition topicPartition : this.offsetAndMetadataMap.keySet()) {
            if (!ownPartitionIdList.contains(topicPartition.partition())) {
                // 分区再平衡时,当前消费者已经不再拥有当前 partition 的消费权
                // 避免出现覆盖,故需要在当前 Consumer 的 offsetAndMetadataMap 移除 当前 partition 的信息
                // 出现场景可以参阅项目目录:  test-kafka-rebalance-log-error.html
                this.offsetAndMetadataMap.remove(topicPartition);
            }
        }
        System.out.printf("%s after: resetConsumerOffsetAndMetadataMap %s %n", Thread.currentThread().getName(), this.offsetAndMetadataMap);
    }
}
```

### ConsumerWorker:  再均衡触发源

```java
package cn.dankal.test.rebalance;

import cn.dankal.test.boot.common.Const;
import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.map.MapUtil;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.serialization.StringDeserializer;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ConcurrentHashMap;

/**
 * 测试 Kafka Rebalance: ConsumerWorker
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class ConsumerWorker {
    /**
     * Map<TopicPartition, OffsetAndMetadata> 当前消费者分区与偏移量信息
     */
    private final Map<TopicPartition, OffsetAndMetadata> offsetAndMetadataMap;

    /**
     * current ConsumerWorker KafkaConsumer instance
     */
    private final KafkaConsumer<String, String> kafkaConsumer;

    /**
     * 是否是临时工作者: 临时工作者在消费五条消息之后将退出当前消费群组
     */
    private final boolean isTempWorker;

    /**
     * 临时工作者的消费消息的次数: 5
     */
    private final static int TEMP_WORK_COUNT = 5;

    /**
     * KafkaConsumer config properties
     */
    private static final Properties PROPERTIES;


    //TODO maybe 事务类 提供事务支持
    //private final Transactional transactional;

    static {
        PROPERTIES = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        PROPERTIES.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 反序列化器
        PROPERTIES.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());
        // message value 反序列化器
        PROPERTIES.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, StringDeserializer.class.getCanonicalName());
        PROPERTIES.put(ConsumerConfig.GROUP_ID_CONFIG, Const.Kafka.ConsumerGroupId.TEST_KAFKA_REBALANCE);

    }

    public ConsumerWorker(boolean isTempWorker) {
        this.offsetAndMetadataMap = new ConcurrentHashMap<>();
        this.isTempWorker = isTempWorker;
        // 消费者线程不安全,建议一个线程一个对应一个消费者实例
        this.kafkaConsumer = new KafkaConsumer<>(PROPERTIES);

        // 设置 分区再均衡监听器
        this.kafkaConsumer.subscribe(
                CollectionUtil.newArrayList(Const.Kafka.Topic.HELLO_TOPIC),
                new RebalanceListener(offsetAndMetadataMap, kafkaConsumer)
        );
    }

    /**
     * 执行 Kafka message consumer work
     */
    public void doWork() {
        // 执行计数 ,搭配 isTempWorker 使用
        int execCount = 0;
        try {
            while (true) {
                ConsumerRecords<String, String> consumerRecords = kafkaConsumer.poll(Duration.ofMillis(500));
                for (ConsumerRecord<String, String> consumerRecord : consumerRecords) {
                    try {
                        // TODO Transaction: transaction begin
                        // transaction.begin();
                        // TODO 消费消息 处理业务
                        long currentMessageOffset = consumerRecord.offset();
                        System.out.printf(
                                "%s 消息消息 ==> 主题: %s 分区: %s 偏移量: %s key: %s value: %s %n",
                                Const.FULL_DATE_TIME_FORMATTER.format(LocalDateTime.now()),
                                consumerRecord.topic(),
                                consumerRecord.partition(),
                                currentMessageOffset,
                                consumerRecord.key(),
                                consumerRecord.value());
                        // TODO  保存分区与偏移量信息
                        TopicPartition topicPartition = new TopicPartition(
                                consumerRecord.topic(),
                                consumerRecord.partition()
                        );
                        OffsetAndMetadata offsetAndMetadata = new OffsetAndMetadata(currentMessageOffset + 1, "no metadata");
                        offsetAndMetadataMap.put(topicPartition, offsetAndMetadata);

                        //TODO 保存偏移量与分区信息
                        if (MapUtil.isNotEmpty(offsetAndMetadataMap)) {
                            for (TopicPartition partition : offsetAndMetadataMap.keySet()) {
                                Const.PARTITION_OFFSET_INFO_MAP.put(
                                        partition,
                                        offsetAndMetadataMap.get(partition).offset()
                                );
                            }
                        }

                        // TODO Transaction: transaction commit
                        // transaction.commit();
                        kafkaConsumer.commitAsync();

                        //TODO 执行计数
                        execCount++;

                        // TODO just test kafka Rebalance
                        if (isTempWorker) {
                            if (execCount == TEMP_WORK_COUNT) {
                                // 触发分区再均衡
                                // 临时工作者工作完成~ 即将推出 当前消费群组
                                System.out.printf(
                                        "%s 完成临时工作即将推出工作群组. offsetAndMetadataMap: %s%n",
                                        Thread.currentThread().getName(),
                                        offsetAndMetadataMap);
                                //提交自己已经完成的工作消息偏移量
                                kafkaConsumer.commitSync();
                                return;
                            }
                        }
                    } catch (Exception e) {
                        // TODO Transaction: transaction rollback
                        // transaction.rollback();

                        // TODO handler exception
                        e.printStackTrace();
                    }
                }
            }
        } finally {
            kafkaConsumer.commitSync();
            kafkaConsumer.close();
        }

    }
}
```

### ProducerWorkerTest:  Kafka 消息提供者

```java
package cn.dankal.test.rebalance;

import cn.dankal.test.boot.common.Const;
import cn.hutool.core.util.RandomUtil;
import org.apache.kafka.clients.producer.KafkaProducer;
import org.apache.kafka.clients.producer.ProducerConfig;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.StringSerializer;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class ProducerWorkerTest {
    /**
     * KafkaProducer
     */
    private static KafkaProducer<String, String> kafkaProducer;


    @BeforeAll
    public static void before() {
        Properties properties = new Properties();
        final String bootstrapServers = "192.168.128.3:9092,192.168.128.3:9093,192.168.128.3:9094";
        // broker(Kafka服务器) 地址清单
        properties.put(ProducerConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers);
        // message key 序列化器
        properties.put(ProducerConfig.KEY_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());
        // message value 序列化器
        properties.put(ProducerConfig.VALUE_SERIALIZER_CLASS_CONFIG, StringSerializer.class.getCanonicalName());

        // 通过 Properties 配置 实例化 KafkaProducer
        kafkaProducer = new KafkaProducer<>(properties);
    }

    /**
     * Kafka 生产者 : 用于 Kafka 分区再均衡的测试
     */
    @Test
    public void testKafkaFirstProducer() throws ExecutionException, InterruptedException {
        for (; ; ) {
            // 消息key (可以为空)
            final String key = "dankal-hello-" + RandomUtil.randomInt(100, 50000000);
            // 消息内容
            final String message = RandomUtil.randomString(64);
            ProducerRecord<String, String> stringProducerRecord =
                    new ProducerRecord<String, String>(
                            Const.Kafka.Topic.HELLO_TOPIC,
                            key,
                            message
                    );
            TimeUnit.MILLISECONDS.sleep(300);
            kafkaProducer.send(stringProducerRecord).get();
        }
    }

    @AfterAll
    public static void after() {
        if (Objects.nonNull(kafkaProducer)) {
            kafkaProducer.close();
        }
    }
}
```

### KafkaRebalanceTest:	测试用例启动类

```java
package cn.dankal.test.rebalance;

import cn.dankal.test.factory.ThreadExecutorFactory;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * 测试 Kafka 分区再均衡
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-26
 */
public class KafkaRebalanceTest {
    /**
     * 测试 Kafka 分区再均衡
     */
    @Test
    public void testKafkaRebalance() throws InterruptedException {
        final int workerCount = 2;
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("dankal-test-kafka-rebalance");
        for (int i = 0; i < workerCount; i++) {
            ConsumerWorker consumerWorker = new ConsumerWorker(false);
            threadPoolExecutor.execute(consumerWorker::doWork);
        }
        TimeUnit.SECONDS.sleep(30);

        //启动一个临时消费者,触发分区再均衡
        ConsumerWorker tempConsumerWorker = new ConsumerWorker(true);
        threadPoolExecutor.execute(tempConsumerWorker::doWork);

        new CountDownLatch(1).await();
    }
}
```

## 日志小析

### 日志全文

[test-kafka-rebalance-log](https://res.mercymodest.com/res/test-kafka-rebalance-log.html)

### 小析

#### 刚开始启动

> dankal-test-kafka-rebalance-12:  [dankal-test-hello-kafka-0, dankal-test-hello-kafka-1]
> dankal-test-kafka-rebalance-12 已完成参与分区再均衡.偏移量信息为: {}
>
> dankal-test-kafka-rebalance-13:  [dankal-test-hello-kafka-2]
> dankal-test-kafka-rebalance-13 已完成参与分区再均衡.偏移量信息为: {}

#### 临时消费者加入 触发分区再平衡

##### onPartitionsRevoked

> dankal-test-kafka-rebalance-12: [dankal-test-hello-kafka-0, dankal-test-hello-kafka-1]
>
> dankal-test-kafka-rebalance-13: [dankal-test-hello-kafka-2]
>
> 
>
> {dankal-test-hello-kafka-2=293, dankal-test-hello-kafka-0=282, dankal-test-hello-kafka-1=299}

##### 分区偏移量信息

> 分区0 消费偏移量：281
>
> 分区1 消费偏移量：298
>
> 分区2 消费偏移量：292

##### onPartitionsAssigned

> dankal-test-kafka-rebalance-12:  [dankal-test-hello-kafka-0]
>
> dankal-test-kafka-rebalance-13:  [dankal-test-hello-kafka-1]
>
> dankal-test-kafka-rebalance-16:  [dankal-test-hello-kafka-2]

##### 分区偏移量信息

> 分区0 消费偏移量：282
>
> 分区1 消费偏移量：299
>
> 分区2 消费偏移量：293

#### 临时消费者退出消费,再次触发分区再平衡

##### 临时消费信息

> dankal-test-kafka-rebalance-16:  [dankal-test-hello-kafka-2]
>
> 最后消费偏移量: 297 [dankal-test-hello-kafka-2]

##### onPartitionsRevoked

> dankal-test-kafka-rebalance-16:  [dankal-test-hello-kafka-2]
>
> 
>
> dankal-test-kafka-rebalance-12:  [dankal-test-hello-kafka-0]
>
> dankal-test-kafka-rebalance-13:  [dankal-test-hello-kafka-1]
>
> 
>
> {dankal-test-hello-kafka-2=298, dankal-test-hello-kafka-0=284, dankal-test-hello-kafka-1=301}

##### 分区偏移量信息

> 分区0 消费偏移量：283
>
> 分区1 消费偏移量：300
>
> 分区2 消费偏移量：297

##### onPartitionsAssigned

> dankal-test-kafka-rebalance-12:   [dankal-test-hello-kafka-0, dankal-test-hello-kafka-1]
>
> dankal-test-kafka-rebalance-13:   [dankal-test-hello-kafka-2]
>
> {dankal-test-hello-kafka-2=298, dankal-test-hello-kafka-0=284, dankal-test-hello-kafka-1=301}

